@@ -26,135 +26,121 @@ sys.path = copy.copy(oldpath)
 
 
 class SchemaQP:
-    """Schema is a general algorithm for integrating heterogeneous data 
-           modalities. It has been specially designed for multi-modal 
-           single-cell biological datasets, but should work in other contexts too.
-       This version is based on a Quadratic Programming Framework.
+    """Schema is a tool for integrating simultaneously-assayed data modalities
 
-       Source code available at: 
-       https://github.com/rs239/schema 
-
-       It is described in the paper “Schema: A general framework for integrating 
-           heterogeneous single-cell modalities” 
-       https://www.biorxiv.org/content/10.1101/834549v1/
-
-       This class provides a sklearn type fit+transform API for affine 
-           transformations of input datasets such that the transformed data 
-           is in agreement with all the input datasets.
-
-
-      Example
-      --------
-           sqp = SchemaQP( min_desired_corr = 0.9) 
-
-           Fit
-           ---
-           # scobj is a scanpy/anndata obj; 3 secondary datasets each given equal wt
-           sqp.fit(scobj.X, 
-                   [scobj.obs.col1.values, scobj.obs.col2.values, datasetY.values], 
-                   ["categorical", "numeric", "feature_vector", "feature_vector_categorical"]) 
-
-           #or 
-
-           # df is a pd.DataFrame, srs is a pd.Series, -1 means try to disagree
-           sqp.fit( df.values, [srs.values], ['numeric'], [-1]) 
-
-           Transform
-           ----------
-           dnew = sqp.transform(scobj.X) 
-        
-           Fit+Transform
-           -------------
-           dnew = sqp.fit_transform( df.values, [srs.values], ['numeric'], [-1]) 
-    
-#### Parameters
-
-`min_desired_corr`: `float` in [0,1)
-
-    The minimum desired correlation between squared L2 distances in the transformed space
-    and distances in the original space.
-
-
-    RECOMMENDED VALUES: At first, you should try a range of values (e.g., 0.99, 0.90, 0.50).
-                        This will give you a sense of what might work well for your data.
-                        After this, you can progressively narrow down your range.
-                        In typical use-cases of large biological datasets,
-                        high values (> 0.80) will probably work best.
-
-
-`w_max_to_avg`: `float` >1, optional (default: 1000)
-
-     Sets the upper-bound on the ratio of w's largest element to w's avg element.
-     Making it large will allow for more severe transformations.
-
-    RECOMMENDED VALUES: We recommend keeping this parameter at its default value (1000); that keep this constraint 
-                        very loose and ensures that  min_desired_corr remains the binding constraint.
-                        Later, as you get a better sense for the right min_desired_corr values
-                        for your data, you can experiment with this too.
-
-                        To really constrain this, set it in the (1-5] range, depending on
-                        how many features you have.
-
-
-`params`: `dict` of key-value pairs, optional (see defaults below)
-
-     Additional configuration parameters.
-     Here are the important ones:
-       * decomposition_model: "pca" or "nmf" (default=pca)
-       * num_top_components: (default=50) number of PCA (or NMF) components to use
-           when mode=="affine".
-
-     You can ignore the rest on your first pass; the default values are pretty reasonable:
-       * dist_npairs: (default=2000000). How many pt-pairs to use for computing pairwise distances
-           value=None means compute exhaustively over all n*(n-1)/2 pt-pairs. Not recommended for n>5000.
-           Otherwise, the given number of pt-pairs is sampled randomly. The sampling is done
-           in a way in which each point will be represented roughly equally.
-       * scale_mode_uses_standard_scaler: 1 or 0 (default=0), apply the standard scaler
-           in the scaling mode
-       * do_whiten: 1 or 0 (default=1). When mode=="affine", should the change-of-basis loadings
-           be made 1-variance?
-
-
-`mode`: `string` one of {`'affine'`, `'scale'`}, optional (default: `'affine'`)
-
-    Whether to perform a general affine transformation or just a scaling transformation
-
-    * 'scale' does scaling transformations only.
-    * 'affine' first does a mapping to PCA or NMF space (you can specify n_components)
-         It then does a scaling transform in that space and then maps everything back to the
-         regular space, the final space being an affine transformation
-
-    RECOMMENDED VALUES: 'affine' is the default, which uses PCA or NMF to do the change-of-basis.
-                        You'll want 'scale' only in one of two cases:
-                         1) You have some features on which you directly want Schema to compute
-                            feature-weights.
-                         2) You want to do a change-of-basis transform other PCA or NMF. If so, you will
-                            need to do that yourself and then call SchemaQP with the transformed
-                            primary dataset with mode='scale'.
-
-#### Returns
-
-    A SchemaQP object on which you can call fit(...), transform(...) or fit_transform(....).
-
-    """
+    The SchemaQP class provides a sklearn type fit+transform API for constrained
+    affine transformations of input datasets such that the transformed data is
+    in agreement with all the input datasets.
+"""
 
     
-    def __init__(self, min_desired_corr=0.99, w_max_to_avg=1000, params={}, mode="affine"):
-        if w_max_to_avg is not None and w_max_to_avg <= 1: raise ValueError("'w_max_to_avg' must be either None or greater than 1")
+    def __init__(self, min_desired_corr=0.99, mode="affine", params={}):
+        """
+
+        :param min_desired_corr: 
+            This parameter controls the severity of the primary modality's
+            transformation, specifying the minimum required correlation
+            between distances in the original space and those in the
+            transformed space. It thus controls the trade-off between
+            deviating further away from the primary modality's original
+            representation and achieving greater agreement with the
+            secondary modalities. Values close to one result in lower
+            distortion of the primary modality while those close to zero
+            enable transformations offering greater agreement
+            between the modalities.
+
+            RECOMMENDED VALUES: In typical single-cell use cases, high
+            values (> 0.80) will probably work best. With these, the
+            distortion will be low, but still be enough for Schema to
+            extract relevant information from the secondary modalities.
+            Furthermore, the feature weights computed by Schema should
+            still be quite infromative.
+
+            The default value of 0.99 is a safe choice to start with; it
+            poses low risk of deviating too far from the primary modality.
+
+            Later, you can experiment with a range of values (e.g., 0.95
+            0.90, 0.80), or use feature-weights aggregated across an
+            ensemble of choices. Alternatively, you can use
+            cross-validation to identify the best setting
+
+        :type min_desired_corr: float in [0,1)
+
+        :param mode: 
+            Whether to perform a general affine transformation or just a
+            scaling transformation
+
+            * `affine` first does a mapping to PCA or NMF space (you can
+              specify num_top_components via the `params` argument). Schema does
+              a scaling transform in the mapped space and then converts everything
+              back to the regular space. The final result is thus an affine
+              transformation in the regular space.
+
+            * `scale` does not do a PCA or NMF mapping, and directly applies
+              the scaling transformation. **Note**: This can be slow if
+              the primary modality's dimensionality is over 100.
+
+
+            RECOMMENDED VALUES: `affine` is the default. You may need `scale` 
+            only in certain cases:
+
+            * You have a limited number of features on which you directly
+              want Schema to compute feature-weights.
+
+            * You want to do a change-of-basis transform other PCA or NMF.
+              If so, you will need to do that yourself and then call
+              SchemaQP with the transformed primary dataset with
+              mode='scale'.
+
+
+        :type mode: string
+
+        :param params: 
+             Dictionary of key-value pairs, specifying additional
+             configuration parameters. Here are the important ones:
+
+               * `decomposition_model`: "pca" or "nmf" (default=pca)
+
+               * `num_top_components`: (default=50) number of PCA (or NMF)
+                 components to use when mode=="affine". We recommend this
+                 setting be <= 100. Schema's runtime is quadratic in this
+                 number.
+
+             You can ignore the rest on your first pass; the default
+             values are pretty reasonable:
+
+               * `dist_npairs`: (default=2000000). How many pt-pairs to use
+                 for computing pairwise distances. value=None means
+                 compute exhaustively over all n*(n-1)/2 pt-pairs. Not
+                 recommended for n>5000.  Otherwise, the given number of
+                 pt-pairs is sampled randomly. The sampling is done in a
+                 way in which each point will be represented roughly
+                 equally.
+
+               * `scale_mode_uses_standard_scaler`: 1 or 0 (default=0),
+                 apply the standard scaler in the scaling mode
+
+               * `do_whiten`: 1 or 0 (default=1). When mode=="affine",
+                 should the change-of-basis loadings be made 1-variance?
+
+        :type params: dict
+
+        :returns: A SchemaQP object on which you can call fit(...), transform(...) or fit_transform(....).
+"""
         if not (mode in ['scale', 'affine']): raise ValueError("'mode' must be one of ['affine','scale']")
         if not (0 <= min_desired_corr < 1): raise ValueError("'min_desired_corr' must be in the range [0,1)") 
 
         self._mode = mode
-        self._w_max_to_avg = w_max_to_avg
+        self._w_max_to_avg = 1000 
         schema_info ("Flag 456.10  ", self._w_max_to_avg)
         self._min_desired_corr = min_desired_corr
         self._std_scaler = None
         self._decomp_mdl = None
-
+        self._orig_decomp_mdl = None
 
         # defaults
-        self._params = {"decomposition_model": "pca",
-                        "num_top_components": 50, #default is 20
+        self._params = {"decomposition_model": "nmf",
+                        "num_top_components": 20, 
                         "dist_npairs": 2000000,
                         "d0_dist_transform": None,
                         "secondary_data_dist_transform_list": None,
@@ -175,87 +161,165 @@ class SchemaQP:
         if not (self._params['num_top_components'] is None or self._params["num_top_components"] >= 2): raise ValueError('Need num_top_components >= 2')
 
 
+
+    def reset_mincorr_param(self, min_desired_corr):
+        """Reset the min_desired_corr. 
+
+        Useful when you want to iterate over multiple choices of this parameter
+        but want to re-use the computed PCA or NMF change-of-basis transform.
+
+        :param min_desired_corr: 
+            The new value of minimum required correlation between original and transformed distances
+
+        :type min_desired_corr: float in [0,1)
+
+"""
+        if min_desired_corr is None or not(0 <= min_desired_corr < 1): raise ValueError("'min_desired_corr' must be between 0 and 1")
+        self._min_desired_corr = min_desired_corr
+
+
+
+    def reset_maxwt_param(self, w_max_to_avg):
+        """ Reset the w_max_to_avg param
+
+        :type w_max_to_avg: float
+
+        :param w_max_to_avg: 
+            The upper-bound on the ratio of Schema weights (w's) largest
+            element to w's avg element.  Making it large will allow This
+            parameter controls the 'deviation' in feature weights and make it
+            large will allow for more severe transformations.
+
+            **Handle with care:** We recommend keeping this parameter at its
+            default value (1000); that keeps this constraint very loose and
+            ensures that min_desired_corr remains the binding constraint.
+            Later, as you get a better sense for the right min_desired_corr
+            values for your data, you can experiment with this too.  To really
+            constrain this, set it in the (1-5] range, depending on how many
+            features you have.
+
+"""
+        if w_max_to_avg is None or w_max_to_avg <= 1: raise ValueError("'w_max_to_avg' must be either None or greater than 1")
+        self._w_max_to_avg = w_max_to_avg
         
     
-    def fit(self, d, secondary_data_val_list, secondary_data_type_list, secondary_data_wt_list = None, d0 = None, d0_dist_transform=None, secondary_data_dist_transform_list=None):
-        """
-Given the primary dataset 'd' and a list of secondary datasets, fit a linear transformation (d*) of
-   'd' such that the correlation between squared pairwise distances in d* and those in secondary datasets
-    is maximized while the correlation between the primary dataset d and d* remains above
-    min_desired_corr
+        
+    
+    def fit(self, d, secondary_data_val_list, secondary_data_type_list, secondary_data_wt_list = None, secondary_data_dist_kernels=None, d0 = None, d0_dist_transform=None):
+        """Compute the optimal Schema transformation, first performing a
+        change-of-basis transformation if required.
+
+        Given the primary dataset `d` and a list of secondary datasets, fit a
+        linear transformation (`d_new`) such that the correlation between
+        squared pairwise distances in `d_new` and those in secondary datasets is
+        maximized while the correlation between the original `d` and the transformed `d_new`
+        remains above min_desired_corr.
+
+        The first three arguments are required, the next is useful, and
+        the rest should be rarely used.
+
+        :type d: Numpy 2-d `array` or Pandas `dataframe`
+
+        :param d: 
+            The primary dataset (e.g. scanpy/anndata's .X).
+
+            The rows are observations (e.g., cells) and the cols are variables (e.g.,
+            gene expression).  The default distance measure computed is L2:
+            sum((point1-point2)**2). Also see `d0_dist_transform`.
 
 
-#### Parameters
+        :type secondary_data_val_list: list of 1-d or 2-d Numpy arrays or Pandas series, each with same number of rows as `d`
 
-`d`: A numpy 2-d `array`
+        :param secondary_data_val_list: 
+            The secondary datasets you want to align the primary data towards.  
 
-    The primary dataset (e.g. scanpy/anndata's .X).
-    The rows are observations (e.g., cells) and the cols are variables (e.g., gene expression).
-    The default distance measure computed is L2: sum((point1-point2)**2). See d0_dist_transform.
+            Columns in Anndata .obs or .obsm variables work well.
 
 
-`secondary_data_val_list`: `list` of 1-d or 2-d numpy `array`s, each with same number of rows as `d`
+        :type secondary_data_type_list: list of strings
 
-    The secondary datasets you want to align the primary data towards.
-    Columns in scanpy's .obs variables work well (just remember to use .values)
+        :param secondary_data_type_list: 
+            The datatypes of the secondary modalities.
+
+            Each element of the list can be one of `numeric,
+            feature_vector, categorical, feature_vector_categorical`.  
+            The list's length should match the length
+            of secondary_data_val_list
 
 
-`secondary_data_type_list`: `list` of `string`s, each value in {'numeric','feature_vector','categorical', 'feature_vector_categorical'}
+            * `numeric`: one floating-pt value for each observation.  The
+              default distance measure is Euclidean: (point1-point2)^2
+ 
+            * `feature_vector`: a k-dimensional vector for each
+              observation.  The default distance measure is Euclidean:
+              sum_{i}((point1[i]-point2[i])^2)
 
-    The list's length should match the length of secondary_data_val_list
+            * `categorical`: a label for each observation.  The default
+              distance measure checks for equality: 1*(val1!=val2)
 
-    * 'numeric' means you're giving one floating-pt value for each obs.
-          The default distance measure is L2:  (point1-point2)**2
-    * 'feature_vector' means you're giving some multi-dimensional representation for each obs.
-          The default distance measure is L2: sum_{i}((point1[i]-point2[i])**2)
-    * 'feature_vector_categorical' means you're giving some multi-dimensional representation for each obs.
-          Each column can take on categorical values, so the distance between two points is sum_{i}(point1[i]==point2[i])
-    * 'categorical' means that you are providing label information that should be compared for equality.
-          The default distance measure is: 1*(val1!=val2)
+            * `feature_vector_categorical`: a vector of labels for each
+              observation.  Each column can take on categorical values, so
+              the distance between two points is
+              sum_{i}(point1[i]==point2[i])
 
 
-`secondary_data_wt_list`: `list` of `float`s, optional (default: `None`)
+        :type secondary_data_wt_list: list of floats, **optional**
 
-    User-specified wts for each dataset. If 'None', the wts are 1.
-    If specified, the list's length should match the length of secondary_data_wt_list
+        :param secondary_data_wt_list: 
+            User-specified wts for each secondary dataset (default= list of 1's)
 
-    NOTE: you can try to get a mapping that *disagrees* with a dataset_info instead of *agreeing*.
-      To do so, pass in a negative number (e.g., -1)  here. This works even if you have just one secondary
-      dataset
+            If specified, the list's length should match the length of
+            secondary_data_val_list. When multiple secondary modalities are
+            specified, this parameter allows you to control their relative
+            weight in seeking an agreement with the primary.
+
+            **Note**: you can try to get a mapping that *disagrees* with a
+            dataset_info instead of *agreeing*.  To do so, pass in a
+            negative number (e.g., -1) here. This works even if you have
+            just one secondary dataset
+
+        :type secondary_data_dist_kernels: list of functions, **optional**
+
+        :param secondary_data_dist_kernels: 
+            The transformations to apply on
+            secondary dataset's L2 distances before using them for correlations.
+
+            If specified, the length of the list should match that of
+            `secondary_data_val_list`. 
+            Each function should take a non-negative float and return a non-negative float. 
+
+            **Handle with care:** Most likely, you don't need this parameter.
+
+        :type d0: A 1-d or 2-d Numpy array, **optional** 
+
+        :param d0: 
+            An alternative representation of the primary dataset.
+
+            This is useful if you want to provide the primary dataset in two
+            forms: one for transforming and another one for computing pairwise
+            distances to use in the QP constraint; if so, `d` is used for the
+            former, while `d0` is used for the latter.
+            If specified, it should have the same number of rows as `d`.
+
+            **Handle with care:** Most likely, you don't need this parameter.
 
 
-`d0`: A 1-d or 2-d numpy array, same number of rows as 'd', optional (default: `None`)
+        :type d0_dist_transform: float -> float function, **optional**
 
-    An alternative representation of the primary dataset.
+        :param d0_dist_transform: 
+            The transformation to apply on d or d0's L2
+            distances before using them for correlations.
 
-    HANDLE WITH CARE! Most likely, you don't need this parameter.
-    This is useful if you want to provide the primary dataset in two forms: one for transforming and
-    another one for computing pairwise distances to use in the QP constraint; if so, 'd' is used for the
-    former, while 'd0' is used for the latter
+            This function should take a non-negative float as input and return a non-negative float.
 
+            **Handle with care:** Most likely, you don't need this parameter.
 
-`d0_dist_transform`: a function that takes a non-negative float as input and
-                    returns a non-negative float, optional (default: `None`)
-
-
-    HANDLE WITH CARE! Most likely, you don't need this parameter.
-    The transformation to apply on d or d0's L2 distances before using them for correlations.
-
-
-`secondary_data_dist_transform`: `list` of functions, each taking a non-negative float and
-                                 returning a non-negative float, optional (default: `None`)
-
-    HANDLE WITH CARE! Most likely, you don't need this parameter.
-    The transformations to apply on secondary dataset's L2 distances before using them for correlations.
-    If specified, the length of the list should match that of secondary_data_val_list
-
-
-#### Returns:
-
-    None
+        :returns: None
          """
         
         if not (d.ndim==2): raise ValueError('d should be a 2-d array')
+        if not np.isfinite(d).all(): raise ValueError('d should not have any NaN or inf values')
+        
         if not (len(secondary_data_val_list) >0): raise ValueError('secondary_data_val_list can not be empty')
 
         if not (len(secondary_data_val_list)==len(secondary_data_type_list)):
@@ -265,7 +329,7 @@ Given the primary dataset 'd' and a list of secondary datasets, fit a linear tra
             raise ValueError('secondary_data_wt_list should have the same length as secondary_data_val_list')
 
         for i in range(len(secondary_data_val_list)):
-            if not (secondary_data_type_list[i] in ['categorical','numeric','feature_vector']):
+            if not (secondary_data_type_list[i] in ['categorical','numeric','feature_vector', 'feature_vector_categorical']):
                 raise ValueError('{0}-th entry in secondary_data_type_list is invalid'.format(i+1))
              
             if not (secondary_data_val_list[i].shape[0] == d.shape[0]):
@@ -274,7 +338,7 @@ Given the primary dataset 'd' and a list of secondary datasets, fit a linear tra
             if not ((secondary_data_type_list[i]=='categorical' and secondary_data_val_list[i].ndim==1) or
                     (secondary_data_type_list[i]=='numeric' and secondary_data_val_list[i].ndim==1) or
                     (secondary_data_type_list[i]=='feature_vector' and secondary_data_val_list[i].ndim==2) or
-                    (secondary_data_type_list[i]=='feature_vector' and secondary_data_val_list[i].ndim==2)):
+                    (secondary_data_type_list[i]=='feature_vector_categorical' and secondary_data_val_list[i].ndim==2)):
                 raise ValueError('{0}-th entry in secondary_data_val_list does not match specified type'.format(i+1))
 
             
@@ -283,7 +347,7 @@ Given the primary dataset 'd' and a list of secondary datasets, fit a linear tra
             
 
         self._params["d0_dist_transform"] = d0_dist_transform
-        self._params["secondary_data_dist_transform_list"] = secondary_data_dist_transform_list
+        self._params["secondary_data_dist_transform_list"] = secondary_data_dist_kernels
         
         if self._mode=="scale":
             self._fit_scale(d, d0, secondary_data_val_list, secondary_data_type_list, secondary_data_wt_list)
@@ -295,21 +359,19 @@ Given the primary dataset 'd' and a list of secondary datasets, fit a linear tra
             
             
     def transform(self, d):
-        """
-Given a dataset `d`, apply the fitted transform to it
+        """Given a dataset `d`, apply the fitted transform to it
+
+        :type d: Numpy 2-d array
+
+        :param d: 
+            The primary modality data on which to apply the transformation. 
+
+            `d` must have with same number of columns as in `fit(...)`.
+            The rows are observations (e.g., cells) and the cols are variables (e.g., gene expression).
 
 
-#### Parameters
-
-`d`:  a numpy 2-d array with same number of columns as primary dataset `d` in the fit(...)
-
-    The rows are observations (e.g., cells) and the cols are variables (e.g., gene expression).
-
-
-#### Returns
-
- a 2-d numpy array with the same shape as `d`
-         """
+        :returns: a 2-d Numpy array with the same shape as `d`
+ """
         
         if self._mode=="scale":
             if not (d.shape[1] == len(self._wts)): raise ValueError('Number of columns in d is incorrect')
@@ -324,29 +386,70 @@ Given a dataset `d`, apply the fitted transform to it
         
 
         
-    def fit_transform(self, d, secondary_data_val_list, secondary_data_type_list, secondary_data_wt_list = None, d0 = None, d0_dist_transform=None, secondary_data_dist_transform_list=None):
+    def fit_transform(self, d, secondary_data_val_list, secondary_data_type_list, secondary_data_wt_list = None, secondary_data_dist_kernels = None, d0 = None, d0_dist_transform=None):
         """
         Calls fit(..) with exactly the arguments given; then calls transform(d).
         See documentation for fit(....) and transform(...) respectively.
 
         """
 
-        self.fit(d, secondary_data_val_list, secondary_data_type_list, secondary_data_wt_list, d0, d0_dist_transform, secondary_data_dist_transform_list)
+        self.fit(d, secondary_data_val_list, secondary_data_type_list, secondary_data_wt_list, secondary_data_dist_kernels, d0, d0_dist_transform)
         return self.transform(d)
 
 
+
+    def feature_weights(self):
+        """
+        Return the feature weights computed by Schema
+
+        If SchemaQP was initialized with `mode=scale`, the weights
+        returned are directly the weights from the quadratic programming
+        (QP), with a weight > 1 indicating the feature was up-weighted.
+
+
+        However, if `mode=affine` was used, the QP-computed weights
+        correspond to columns of the PCA or NMF decomposition. In that
+        case, this functions maps them back to the primary dataset's
+        features. This is done by a 'softmax'-type summation of loadings
+        across the PCA/NMF columns, with each column's weight proportional
+        to exp(QP wt). 
+
+        If you'd like to build your own mapping from
+        PCA/NMF weights to gene weights, look at the code for this
+        function. In particular, you'll need the QP weights (in `self._wts`( and the
+        PCA/NMF model (in `self._decomp_mdl`)
+
+        *returns* : a vector of floats, the same size as the primary dataset's dimensionality
+"""
+        if self._mode == "scale":
+            return self._wts
+        else:
+            w = np.exp(self._wts)  # exponentiation of wts so the highest-wt features really stand out
+            w[ self._wts <= 1] = 0 # ignore features with wt <=1 (these were not up-weighted)
+            w = w/np.sum(w) # normalize
+
+            if self.params['decomposition_model'] == 'nmf':
+                feat_wts = sum([ w[i]*self._decomp_mdl.components_[i,:] for i in range(len(w))])
+                
+            else: #in PCA, loadings are unique upto a sign, so when adding them up, we just take the magnitudes 
+                feat_wts = sum([ w[i]*np.abs(self._decomp_mdl.components_[i,:]) for i in range(len(w))])
+
+            return feat_wts
+
+        
+
     ###################################################################
     ######## "private" methods below. Not that Python cares... ########
     ###################################################################
     def _getDistances(self, D, d0_in, G, nPointPairs):
         """
         Compute the various distances between point pairs in D
-        D is a NxK numpy type 2-d array, with N points, each of K dimensions
+        D is a NxK Numpy type 2-d array, with N points, each of K dimensions
         d0_in could be None, in which D is used to compute distances for the original space.
            Otherwise, d0_in is Nxr (r>=1) array and original-space distances are computed from this. 
         G is list, with each entry a 3-tuple: (g_val, g_type, gamma_g)
            each 3-tuple corresponds to one dimension of side-information
-             g_val is Nx1 numpy type 1-d vector of values for the N points, in the same order as D
+             g_val is Nx1 Numpy type 1-d vector of values for the N points, in the same order as D
              g_type is "numeric", "categorical", or "feature_vector"
              gamma_g is the relative wt you want to give this column. You can leave 
                it as None for all 3-tuples, but not for just some. If you leave them
@@ -708,42 +811,63 @@ Given a dataset `d`, apply the fitted transform to it
         print('Running quadratic program...', end='', flush=True)
         s, sl = self._fit_helper(dx, d0, secondary_data_val_list, secondary_data_type_list, secondary_data_wt_list)
         
-        self._wts = np.maximum(s,0)
+        self._wts = np.maximum(s,0) # shouldn't be <0 anyways, except for numerical issues
+        self._wts = len(self._wts)*self._wts/np.sum(self._wts) #let's normalize so that avg wt = 1
+        
         self._soln_info = dict(sl)
         print(' done.\n', end='', flush=True)
                 
+
         
     def _fit_affine(self, d, d0, secondary_data_val_list, secondary_data_type_list, secondary_data_wt_list):
         do_whiten = self._params.get("do_whiten",1)>0
         ncomp = self._params.get("num_top_components",None) #default is all
-
-        dx1 = d.copy()
         model_type = self._params.get("decomposition_model","pca").lower()
-
-        print('Running change-of-basis transform ({0}, {1} components)...'.format(model_type, ncomp), end='', flush=True)
+        dx1 = d.copy()
         
-        if model_type=="pca":
-            if not scipy.sparse.issparse(dx1):
-                self._decomp_mdl = sklearn.decomposition.PCA(n_components=ncomp, whiten=do_whiten)
-            else:
-                self._decomp_mdl = sklearn.decomposition.TruncatedSVD(n_components=ncomp)
-                schema_warning("Using TruncatedSVD instead of PCA because input is a sparse matrix. do_whiten arguments will be ignored")
-            dx = self._decomp_mdl.fit_transform(dx1)
+        if self._decomp_mdl is None:
+            print('Running change-of-basis transform ({0}, {1} components)...'.format(model_type, ncomp), end='', flush=True)
+
+            if model_type=="pca":
+                if not scipy.sparse.issparse(dx1):
+                    self._decomp_mdl = sklearn.decomposition.PCA(n_components=ncomp, whiten=do_whiten)
+                else:
+                    self._decomp_mdl = sklearn.decomposition.TruncatedSVD(n_components=ncomp)
+                    schema_warning("Using TruncatedSVD instead of PCA because input is a sparse matrix. do_whiten arguments will be ignored")
+                    
+                self._decomp_mdl.fit(dx1)
+
+            elif model_type=="nmf":
+                self._decomp_mdl = sklearn.decomposition.NMF(n_components=ncomp, init=None)
+                W = self._decomp_mdl.fit(dx1)
+                
+            self._orig_decomp_mdl = copy.deepcopy(self._decomp_mdl)
+            print(' done.')
             
+        else:
+            print('Reusing previous change-of-basis transformation')
+            self._decomp_mdl = copy.deepcopy(self._orig_decomp_mdl)
+
+        if model_type=="pca":
+            dx = self._decomp_mdl.transform(dx1)
+
         elif model_type=="nmf":
-            self._decomp_mdl = sklearn.decomposition.NMF(n_components=ncomp, init=None)
-            W = self._decomp_mdl.fit_transform(dx1, W=None, H=None)
+            W = self._decomp_mdl.transform(dx1)
             if do_whiten:
                 H = self._decomp_mdl.components_
                 wsd = W.std(axis=0)
                 W = W/wsd
                 self._decomp_mdl.components_ *= wsd[:,None]
             dx = W
+
+
             
-        print(' done.\nRunning quadratic program...', end='', flush=True)
+        print('Running quadratic program...', end='', flush=True)
         s, sl = self._fit_helper(dx, d0, secondary_data_val_list, secondary_data_type_list, secondary_data_wt_list)
 
-        self._wts = np.maximum(s,0)
+        self._wts = np.maximum(s,0) # shouldn't be <0 anyways, except for numerical issues
+        self._wts = len(self._wts)*self._wts/np.sum(self._wts) #let's normalize so that avg wt = 1
+
         self._soln_info = dict(sl)
         print(' done.\n', end='', flush=True)
 
